///////////////////////////////////////////////////////////////////////////////
// statistics.C: routines for statistical analysis of AuxFields.
//
// This version, modified for use with scat, differs from standard semtex
// because it has to accumlate correlations for passive scalar.
//
// Copyright (c) 1994 <--> $Date: 2013/10/21 20:52:12 $, Hugh Blackburn
//
// The collection of statistics is controlled by the setting of the
// AVERAGE token. Legal values are 0 (default), 1, 2. The routines
// here do not control how often statistics are updated: that happens
// in Analyser class methods.
//
// All collected statistics are given 1-character names. Potentially
// these may clash with quantities derived/used by addfield utility
// for example. Caveat emptor. In the longer term we should move to
// CSV for field names.
//
// AVERAGE = 0.  No statistics.
//
// AVERAGE = 1.  Running averages of variables held by the Domain used
// for initialisation. The data are held in semi-Fourier state.
//
// AVERAGE = 2.  Additionally, correlation terms for computation of
// Reynolds stresses. (Correlations, based on products of variables,
// are computed and held in physical space.)
// 
// Naming scheme for components of the symmetric "Reynolds stresses" tensor:
//
//                      / uu uv uc uw \     /  A  B  D  G \
//                      | .  vv vc vw |  =  |  .  C  E  H |
//                      | .  .  cc cw |     |  .  .  F  I |
//                      \ .  .  .  ww /     \  .  .  .  J /
//
// Note the structure is different/names conflict compared to the
// plain non-scalar case. Even for 2C there will be 6 correlations,
// and 10 for 3C. Fortunately the code modifications required are
// fairly minimal. Note also that we have dropped AVERAGE = 3, or
// correlations for the energy balance.
//
///////////////////////////////////////////////////////////////////////////////

static char RCS[] = "$Id: statistics.cpp,v 1.1 2013/10/21 20:52:12 hmb Exp $";

#include <sem.h>


Statistics::Statistics (Domain* D) :
// ---------------------------------------------------------------------------
// Store averages for all Domain Fields, and correlations.
// ---------------------------------------------------------------------------
  _name (D -> name),
  _base (D),
  _iavg (Femlib::ivalue ("AVERAGE")),
  _nraw (_base -> nField()),
  _nvel (_nraw - 1),
  _nrey ((_iavg > 1) ? ((_nvel+1)*_nvel)/2 : 0),
  _neng (0)
{
  if (_iavg == 0) return;
  if ((_iavg  < 0) || (_iavg > 2))
    message ("Statistics::Statistics", "AVERAGE token out of [0,2]", ERROR);
					 
  int_t       i;
  const int_t nz   = Geometry::nZProc();
  const int_t ntot = Geometry::nTotProc();

  // -- Set pointers, allocate storage.

  for (i = 0; i < _nraw; i++)	// -- Local pointers to raw variables.
    _raw[_base -> u[i] -> name()] = (AuxField*) _base -> u[i];

  if (_iavg > 0) // -- Set up buffers for averages of raw variables.
    for (i = 0; i < _nraw; i++)
      _avg[_base -> u[i] -> name()] =
	new AuxField (new real_t[ntot], nz, _base->elmt, _base->u[i]->name());

  if (_iavg > 1) // -- Set up buffers for Reynolds stress correlations.
    for (i = 0; i < _nrey; i++)
      _avg['A' + i] = new AuxField (new real_t[ntot], nz, _base->elmt, 'A'+i);
}


void Statistics::initialise (const char* filename)
// ---------------------------------------------------------------------------
// This is for standard running averages. Try to initialize from file
// filename (e.g. "session.avg"), failing that set all buffers to
// zero.  Number of fields in file should be same as Statistics::_avg
// buffer.
// ---------------------------------------------------------------------------
{
  ROOTONLY cout << "-- Initialising averaging  : ";  

  ifstream file (filename);
  map<char, AuxField*>::iterator k;

  if (file) {
    ROOTONLY {
      cout << "read from file " << filename;
      cout.flush();
    }
    file >> *this;
    file.close();

    // -- Fourier transform raw data components.

    for (k = _raw.begin(); k != _raw.end(); k++)
      _avg[k -> second -> name()] -> transform (FORWARD);
  
  } else {			// -- No file, set to zero.
    ROOTONLY cout << "set to zero";
    for (k = _avg.begin(); k != _avg.end(); k++) *(k -> second) = 0.0;
    _navg = 0;
  }

  ROOTONLY cout << endl;
}


void Statistics::update (AuxField** wrka,
			 AuxField** wrkb)
// ---------------------------------------------------------------------------
// Update running averages, using arrays wrka & wrkb as workspace.
// All product/correlation terms are calculated without dealiasing,
// and are held in physical space.
// ---------------------------------------------------------------------------
{
  if (_iavg < 1) return;
  
  int_t  i, j;
  Field* master = _base -> u[0];
  map<char, AuxField*>::iterator k;

  // -- Weight old running averages.

  for (k = _avg.begin(); k != _avg.end(); k++)
    *(k -> second) *= static_cast<real_t>(_navg);
    
  // -- Always do running averages of raw data.

  for (k = _raw.begin(); k != _raw.end(); k++)
    *_avg[k -> second-> name()] += *(k -> second);

  // -- Reynolds stress correlations.
  //    After this, wrka contains current velocity data in physical space.

  if (_iavg > 1) {
    (*wrka[0] = *_raw['u']) . transform (INVERSE);
    (*wrka[1] = *_raw['v']) . transform (INVERSE);
    (*wrka[2] = *_raw['c']) . transform (INVERSE);
    if (_nvel == 4) (*wrka[3] = *_raw['w']) . transform (INVERSE);

    _avg['A'] -> timesPlus (*wrka[0], *wrka[0]);
    _avg['B'] -> timesPlus (*wrka[0], *wrka[1]);
    _avg['C'] -> timesPlus (*wrka[1], *wrka[1]);
    _avg['D'] -> timesPlus (*wrka[0], *wrka[2]);
    _avg['E'] -> timesPlus (*wrka[1], *wrka[2]);
    _avg['F'] -> timesPlus (*wrka[2], *wrka[2]);

    if (_nvel == 4) {
      _avg['G'] -> timesPlus (*wrka[0], *wrka[3]);
      _avg['H'] -> timesPlus (*wrka[1], *wrka[3]);
      _avg['I'] -> timesPlus (*wrka[2], *wrka[3]);
      _avg['J'] -> timesPlus (*wrka[3], *wrka[3]);
    }
  }

  // -- Normalise and smooth running averages.

  for (k = _avg.begin(); k != _avg.end(); k++) {
    master -> smooth (k -> second);
    *(k -> second) /= static_cast<real_t>(_navg + 1);
  }

  _navg++;
}


void Statistics::dump (const char* filename)
// ---------------------------------------------------------------------------
// Similar to Domain::dump.
//
// As of 24/11/2004, we deleted the checkpointing that used to happen:
// all dumping now happens to file named on input.
//
// We also smooth all the outputs with the mass matrix.
// ---------------------------------------------------------------------------
{
  const int_t step     = _base -> step;
  const bool  periodic = !(step %  Femlib::ivalue ("IO_FLD"));
  const bool  initial  =   step == Femlib::ivalue ("IO_FLD");
  const bool  final    =   step == Femlib::ivalue ("N_STEP");

  if (!(periodic || final)) return;

  ofstream    output;
  int_t       i;
  map<char, AuxField*>::iterator k;

  ROOTONLY {
    const char routine[] = "Statistics::dump";
    const bool verbose   = static_cast<bool> (Femlib::ivalue ("VERBOSE"));

    output.open (filename);
    if (!output) message (routine, "can't open dump file", ERROR);
    if (verbose) message (routine, ": writing field dump", REMARK);
  }
  
  // -- All terms are written out in physical space but some are
  //    held internally in Fourier space.

  for (k = _raw.begin(); k != _raw.end(); k++)
    _avg[k -> second -> name()] -> transform (INVERSE);

  output << *this;

  for (k = _raw.begin(); k != _raw.end(); k++)
    _avg[k -> second -> name()] -> transform (FORWARD);

  ROOTONLY output.close();
}


ofstream& operator << (ofstream&   strm,
		       Statistics& src )
// ---------------------------------------------------------------------------
// Output Statistics class to file.  Like similar Domain routine.
// ---------------------------------------------------------------------------
{
  int_t             i;
  const int_t       N = src._avg.size();
  vector<AuxField*> field (N);
  map<char, AuxField*>::iterator k;

  for (i = 0, k = src._avg.begin(); i < N; i++, k++) field[i] = k -> second;

  writeField (strm, src._name, src._navg, src._base -> time, field);

  return strm;
}


ifstream& operator >> (ifstream&   strm,
		       Statistics& tgt )
// ---------------------------------------------------------------------------
// Input Statistics class from file.  Like similar Domain routine.
// ---------------------------------------------------------------------------
{
  const char routine[] = "strm>>Statistics";
  int_t i, j, np, nz, nel, ntot, nfields;
  int_t npchk,  nzchk, nelchk;
  char  s[StrMax], f[StrMax], err[StrMax], fields[StrMax];
  bool  swap = false;
  map<char, AuxField*>::iterator k;

  if (strm.getline(s, StrMax).eof()) return strm;
  
  strm.getline (s, StrMax) . getline (s, StrMax);

  string ss(s);
  istringstream sss (ss);
  sss >> np >> np >> nz >> nel;
 
  tgt._avg.begin()->second->describe (f);
  sss.clear ();
  sss.str   (ss = f);
  sss >> npchk >> npchk >> nzchk >> nelchk;

  if (np  != npchk ) message (routine, "element size mismatch",       ERROR);
  if (nz  != nzchk ) message (routine, "number of z planes mismatch", ERROR);
  if (nel != nelchk) message (routine, "number of elements mismatch", ERROR);
  
  ntot = np * np * nz * nel;
  if (ntot != Geometry::nTot())
    message (routine, "declared sizes mismatch", ERROR);

  strm.getline (s, StrMax);

  sss.clear();
  sss.str  (ss = s);
  sss >> tgt._navg;
    
  strm.getline (s, StrMax) . getline (s, StrMax);
  strm.getline (s, StrMax) . getline (s, StrMax) . getline (s, StrMax);
    
  nfields = 0;
  while (isalpha (s[nfields])) {
    fields[nfields] = s[nfields];
    nfields++;
  }
  fields[nfields] = '\0';

#if 0
  if (nfields != tgt._avg.size()) {
    sprintf (err, "strm: %1d fields, avg: %1d", nfields, 
	     static_cast<int_t>(tgt._avg.size()));
    message (routine, err, ERROR);
  }
#endif

  for (i = 0, k = tgt._avg.begin(); k != tgt._avg.end(); k++, i++)
    if (!strchr (fields, k -> second -> name())) {
      sprintf (err, "field %c not present in avg, start with 0.0", 
	       k -> second -> name());
      message (routine, err, WARNING);
      *k -> second = 0.0;
    }

  strm.getline (s, StrMax);
  Veclib::describeFormat (f);

  if (!strstr (s, "binary"))
    message (routine, "input field strm not in binary format", ERROR);
  
  if (!strstr (s, "endian"))
    message (routine, "input field strm in unknown binary format", WARNING);
  else {
    swap = ((strstr (s, "big") && strstr (f, "little")) ||
	    (strstr (f, "big") && strstr (s, "little")) );
    ROOTONLY {
      if (swap) cout << " (byte-swapping)";
      cout.flush();
    }
  }

  for (j = 0; j < nfields; j++) {
    strm >> *tgt._avg[fields[j]];
    if (swap) tgt._avg[fields[j]] -> reverse();
  }
  
  ROOTONLY if (strm.bad())
    message (routine, "failed reading average file", ERROR);

  return strm;
}


void Statistics::phaseUpdate (const int_t j   ,
			      AuxField**  wrka,
			      AuxField**  wrkb)
// ---------------------------------------------------------------------------
// Phase updates are running updates, like those for standard
// statistics, but they are only computed at (a presumed very limited)
// number of instants per run. And so there would be a number of
// averaging files, to be called e.g. session.0.phs, session.1.phs
// ... session.(N-1).phs. Instead of reserving enough memory for all
// these buffers, we just keep workspace reserved, and at every phase
// point the appropriate file (number j) is read in, updated, and
// written back out. If the file does not exist, it is created.
//
// NB: the number of averages computed (needed for the running
// averaging) is only updated when j == 0, which will happen as the
// last of a set of phase averages.
// ---------------------------------------------------------------------------
{
  char filename[StrMax];
  sprintf (filename, "%s.%1d.phs", _name, j);

  this -> initialise (filename);
  this -> update     (wrka, wrkb);
  this -> dump       (filename);
}
